#!/usr/bin/ruby
# coding: utf-8

# おおよそランダムなグラフを作るスクリプト．
# 引数は3つ．順にノード数(n)，次数(d)，距離制約(r)．
# 実行例: ruby randomGridRegular.rb 16 3 2

if ARGV.length == 0
  puts "Usage: $0 order degree length"
  exit
end

@n = ARGV.shift.to_i
@nx = @ny = Math.sqrt(@n).to_i
@d = ARGV.shift.to_i
@r = ARGV.shift.to_i
@verbose = 0

@prefix = $0.sub(/.rb$/, '')
puts "n = #{@n} = #{@nx} x #{@ny}, d = #{@d}, r = #{@r}"

# 平面座標 xy = [x, y] をノード番号 i に変換する
# 0 <= x, y < @nx, 0 <= i < @n
def xy2i(xy)
  x, y = xy
  x + y * @nx
end

# ノード番号 i を平面座標 xy = [x, y] に変換する
# 0 <= x, y < sqrt(n), 0 <= i < n
def i2xy(i)
  [i % @nx, i / @nx]
end

# ノードiからのマンハッタン距離がr以下のノードを返す
def nodesManhattanDist(i, r)
  js = []
  if i.is_a?(Array)
    xi, yi = i
    puts "#{js}"
    js
  elsif i.is_a?(Integer)
    xi, yi = i2xy(i)
  end
  (- r..r).each { |xr|
    yr0 = r - xr.abs
    (- yr0..yr0).each { |yr|
      next if xr == 0 && yr == 0
      js << [xr + xi, yr + yi] if 0 <= xr + xi && xr + xi < @nx && 0 <= yr + yi && yr + yi < @ny
    }
  }
  if i.is_a?(Array)
    puts "#{i}-#{r}: #{js}" if @verbose > 2
    js
  elsif i.is_a?(Integer)
    js.map! { |x, y| xy2i([x, y]) }
    puts "#{i}-#{r}: #{js}" if @verbose > 1
    js
  end
end

# マンハッタン距離制約 r を満たすグラフを作る．
# 連結か否かは問わない．次数は最大値 d を満たそうとするが気にしない．
# 枝のないノードが残る場合もある
def randomGraph(e)
  (0...@n).to_a.each { |i| # 番号の大きいノードに枝が残りがち
    xi, yi = i2xy(i)
    js = nodesManhattanDist(i, @r) # マンハッタン距離制約 r を満たすノード
    js = js.map { |j| # 次数が @d 以下のノードに絞る
      next if j < i
      if e[j].length < @d
        j
      else
        nil
      end
    }.compact
    js = js.shuffle! # sample(@d - e[i].length) にしていたが，距離が長いものを高頻度で選ぶために，同じ値がなんども出てくるので．
    js.each { |j|
      break if e[i].length >= @d # 枝をすべて貼り終えた．
      next if e[i].index(j) != nil # すでに同じ枝がある．
      e[i] << j
      e[j] << i
    }
    puts "node #{i}: #{e[i]}" if @verbose > 0
  }
  e
end

def regular(e)
  degree = e.map { |ei| ei.length }
  tries = 0
  puts "degrees: #{degree.sort.first(20)}..."
  fadd = true
  while degree.min < @d && fadd && (tries += 1) <= 100
    if true # test
      dmin = degree.min
      dmins = (0...@n).to_a.map { |i|
        if e[i].length == dmin
          i
        else
          nil
        end
      }.compact
      puts "#{dmin} degree nodes: #{dmins}"
    end
    # 適当に増やして，あとで削除
    fadd = false
    (0...@n).to_a.each { |i| # 番号の大きいノードに枝が残りがち
      next if e[i].length >= @d
      js = nodesManhattanDist(i, @r) # マンハッタン距離制約 r を満たすノード
      jss = (0..@n).to_a.map { [] }
      js.each { |j| # jを次数ごとに分ける
        jss[e[j].length] << j
      }
      jss.each { |js1|
        js1.shuffle.each { |j|
          break if e[i].length >= @d
          next if e[i].index(j) != nil
          # p ["add", i, j]
          e[i] << j
          e[j] << i
          fadd = true
        }
      }
    }
    # 次数が超えたノードから枝を削除
    (0...@n).to_a.each { |i| # 番号の大きいノードに枝が残りがち
      next if e[i].length <= @d
      jss = (0..@n).to_a.map { [] }
      e[i].each { |j| # 次数が @d 以下のノードに絞る
        jss[e[j].length] << j
      }
      jss.reverse_each { |js1| # 次数字数が多い方から順に
        if true
          # 中心に近いノードとの枝を削除
          js1 = js1.shuffle.sort { |j1, j2|
            xj1, yj1 = i2xy(j1)
            xj2, yj2 = i2xy(j2)
            c1 = (xj1 - @nx / 2).abs + (yj1 - @nx / 2).abs # およその中心からのマンハッタン距離
            c2 = (xj2 - @nx / 2).abs + (yj2 - @nx / 2).abs
            c1 <=> c2 # 中心に近いものを優先
          }
        else
          # 一様なランダム
          js1.shuffle!
        end
        js1.each { |j|
          break if e[i].length <= @d
          # p ["delete", i, j]
          e[i].delete(j)
          e[j].delete(i)
        }
      }
    }
    degree = e.map { |ei| ei.length }
  end
  puts "degrees (tried: #{tries} times): #{degree.sort.first(20)}..."
  e
end

e = (0...@n).to_a.map { [] }
e = randomGraph(e)
e = regular(e)

# ファイルに保存
fn="#{@prefix}-n#{@n}d#{@d}r#{@r}-#{Process.pid}.edges"
puts "save into \"#{fn}\""
f = File.open(fn, "w")

f.puts "# #{$0} #{@n} #{@d} #{@r}"

# @r を検算
rs = (0..@r + 1).to_a.map { 0 }
(0...e.length).each { |i|
  xi, yi = i2xy(i)
  e[i].each { |j|
    next if i > j
    xj, yj = i2xy(j)
    rs[[(xi - xj).abs + (yi - yj).abs, @r + 1].min] += 1
  }
}
puts "r check: #{rs}"
f.puts "# r: #{rs}"

# 次数の分布
ds = (0..@d).to_a.map { [] }
(0...e.length).each { |i|
  ds[e[i].length] << i
}
puts "degree check: #{ds.map { |di| di.length}}"
f.puts "# degree check: #{ds.map { |di| di.length}}"
(0...@d).each { |d|
  puts "degree #{d} nodes: #{ds[d]} = #{ds[d].map { |i| i2xy(i) }}" if ds[d].length > 0
  f.puts "# degree #{d} nodes: #{ds[d]} = #{ds[d].map { |i| i2xy(i) }}" if ds[d].length > 0
}

(0...@n).each { |i|
  e[i].sort.each { |j|
    next if j < i
    f.puts "#{i} #{j}"
  }
}
f.close
